#include <windows.h>
#include <shlwapi.h>
#include <shlobj.h>
#include <dbt.h>
#include <stdio.h>
#include <conio.h>
#include <errno.h>
#include <intrin.h>
#include "whirlpool\whirlpool.h"
#include "xts\xts.h"
#include "scrypt\scrypt.h"
#include "misc\misc.h"

#define IMDPROXY_FLAG_RO 0x1
#define IMDPROXY_HEADER_SIZE 4096
#define DEF_BUFFER_SIZE (1 << 20)
#define AIM_BUFFER_SIZE (2 << 20)
#define DEF_REQUIRED_ALIGNMENT 1

typedef enum {
	IMDPROXY_REQ_NULL, IMDPROXY_REQ_INFO,
	IMDPROXY_REQ_READ, IMDPROXY_REQ_WRITE,
	IMDPROXY_REQ_CONNECT, IMDPROXY_REQ_CLOSE
} IMDPROXY_REQ;
typedef struct {ULONGLONG file_size, req_alignment, flags;} IMDPROXY_INFO_RESP;
typedef struct {ULONGLONG request_code, offset, length;} IMDPROXY_IO_REQ;
typedef struct {ULONGLONG errorno, length;} IMDPROXY_IO_RESP;

static int buffer_size = DEF_BUFFER_SIZE;
static __int64 offset = 0, extra_offset = 0;
static int sector_size = 512;
static char comm_device[20];
static IMDPROXY_INFO_RESP proxy_info = {0};
static TCHAR file_name[32768 + 200], *backup_file = NULL;
static TCHAR drive_letter[3], *mount_point = NULL;
static DWORD timeout = INFINITE;
static BOOL use_prompt = TRUE;
static BOOL suspend_close = FALSE;
static TCHAR *format_param = L"";
static int verbose = 1;
static char global_str[80];
static TCHAR *password;
static unsigned int pass_size;
static unsigned char header_key_whirlpool[6][32];
static unsigned char header_key_sha3[6][32];
static unsigned char salt_whirlpool[64 + 4], salt_sha3[64 + 4];
static BOOL detach_console = FALSE;
static unsigned long enc_cipher = 0, dec_cipher = 0;
static SYSTEM_INFO si;
static HANDLE encrypt_event[32], decrypt_event[32];
static HANDLE encrypt_done[32], decrypt_done[32];
static void *volatile thread_buffer_read, *volatile thread_buffer_write;
static volatile int request_size_read, request_size_write;
static volatile __int64 thread_offset_read, thread_offset_write;
static OVERLAPPED overlap_read, overlap_write = {0};
static unsigned char *write_buf;
static int expected_write = 0;
static BOOL force_aim = FALSE, use_imdisk;
static HANDLE h, h_cpl, exit_event, clean_ok;
static FARPROC ImDiskOpenDeviceByMountPoint, ImDiskForceRemoveDevice, ImDiskRemoveMountPoint;
static FILETIME access_time, write_time;

static char *cipher_name[] = {"", "AES", "Serpent", "Twofish"};

#pragma pack(push, 1)
typedef struct {
	unsigned char salt[64]; short version;
	unsigned char hash[WHIRLPOOL_DIGESTSIZE]; unsigned char master_key[6][32];
	__int64 volume_offset; __int64 volume_size;
} HEADER_STRUCT;
#pragma pack(pop)


static void out(char *format, ...)
{
	if (!verbose) return;
	va_list args;
	va_start(args, format);
	if (detach_console) AttachConsole(ATTACH_PARENT_PROCESS);
	vprintf(format, args);
	if (detach_console) FreeConsole();
	va_end(args);
}

__declspec(noinline) static void disp_err(char *txt)
{
	fputs(txt, stderr);
}

/*void disp(void *buf, int size)
{
	int i;
	for (i = 0; i < size; i++) out("%02X", ((unsigned char*)buf)[i]);
	out("\n");
}*/

static DWORD __stdcall derive_pass_whirlpool(LPVOID lpParam)
{
	return scrypt(PRF_WHIRLPOOL, password, pass_size, salt_whirlpool, 64, 32768, 16, 16, header_key_whirlpool, sizeof header_key_whirlpool);
}

static DWORD __stdcall derive_pass_sha3(LPVOID lpParam)
{
	return scrypt(PRF_SHA3, password, pass_size, salt_sha3, 64, 32768, 16, 16, header_key_sha3, sizeof header_key_sha3);
}


static DWORD __stdcall encrypt_thread(LPVOID lpParam)
{
	int i, i_start = (size_t)lpParam << 9, i_inc = si.dwNumberOfProcessors << 9;

	for (;;) {
		WaitForSingleObject(encrypt_event[(int)(size_t)lpParam], INFINITE);
		for (i = i_start; i < request_size_write; i += i_inc)
			XTS_EncryptSector(enc_cipher, (thread_offset_write + i) >> 9, thread_buffer_write + i);
		SetEvent(encrypt_done[(int)(size_t)lpParam]);
	}
	return 0;
}

static DWORD __stdcall decrypt_thread(LPVOID lpParam)
{
	int i, i_start = (size_t)lpParam << 9, i_inc = si.dwNumberOfProcessors << 9;

	for (;;) {
		WaitForSingleObject(decrypt_event[(int)(size_t)lpParam], INFINITE);
		for (i = i_start; i < request_size_read; i += i_inc)
			XTS_DecryptSector(dec_cipher, (thread_offset_read + i) >> 9, thread_buffer_read + i);
		SetEvent(decrypt_done[(int)(size_t)lpParam]);
	}
	return 0;
}


static int physical_read(void *buf, int size, __int64 offset)
{
	int i, current_size, next_size, size_done;
	DWORD bytes_read;
	__int64 *overlap_read_Offset = (__int64*)&overlap_read.Offset;

	next_size = min(size, 32768);
	*overlap_read_Offset = offset;
	ReadFile(h, buf, next_size, NULL, &overlap_read);
	size_done = 0;

	if (verbose >= 2) out(" ? Read: %d bytes at offset %I64d\n", size, offset);

	do {
		current_size = next_size;
		next_size = min(size - (size_done + current_size), 65536);

		thread_buffer_read = buf + size_done;
		request_size_read = current_size;
		thread_offset_read = offset + size_done;
		i = 0;

		GetOverlappedResult(h, &overlap_read, &bytes_read, TRUE);
		if (bytes_read != current_size) {
			fprintf(stderr, "%%Read error: %u / %d bytes at offset %I64d.\n", (unsigned int)bytes_read, current_size, *overlap_read_Offset);
			return EIO;
		}
		if (next_size) {
			*overlap_read_Offset = offset + size_done + current_size;
			ReadFile(h, buf + size_done + current_size, next_size, NULL, &overlap_read);
		}
		do SetEvent(decrypt_event[i]); while (++i < si.dwNumberOfProcessors);
		size_done += current_size;
		WaitForMultipleObjects(si.dwNumberOfProcessors, decrypt_done, TRUE, INFINITE);
	} while (next_size);

	return 0;
}

static int physical_write(void *buf, int size, __int64 offset)
{
	int i, err = 0;
	DWORD bytes_written;
	__int64 *overlap_write_Offset = (__int64*)&overlap_write.Offset;

	thread_buffer_write = buf;
	request_size_write = size;
	thread_offset_write = offset;
	i = 0; do SetEvent(encrypt_event[i]); while (++i < si.dwNumberOfProcessors);
	WaitForMultipleObjects(si.dwNumberOfProcessors, encrypt_done, TRUE, INFINITE);

	GetOverlappedResult(h, &overlap_write, &bytes_written, TRUE);
	if (bytes_written != expected_write) {
		fprintf(stderr, "%%Write error: %u / %d bytes at offset %I64d.\n", (unsigned int)bytes_written, expected_write, *overlap_write_Offset);
		err = EIO;
	}
	memcpy(write_buf, buf, size);
	*overlap_write_Offset = offset;
	WriteFile(h, write_buf, size, NULL, &overlap_write);
	expected_write = size;

	if (verbose >= 2) out(" ! Write: %d bytes at offset %I64d\n", size, offset);

	return err;
}


static DWORD __stdcall aim_dismount(LPVOID lpParam)
{
	HANDLE device;
	char path[256];
	STARTUPINFOA si = {sizeof si};
	PROCESS_INFORMATION pi;

	strcpy(path, "\\\\.\\A:");
	do
		if ((device = CreateFileA(path, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_NO_BUFFERING, NULL)) != INVALID_HANDLE_VALUE) {
			FlushFileBuffers(device);
			CloseHandle(device);
		}
	while (++path[4] <= 'Z');

	sprintf(path, lpParam ? "ArsenalImageMounterControl.exe /remove" :
				  "cmd /v:on /c \"for /f \"tokens=1,3\" %%I in ('ArsenalImageMounterControl.exe /query') do "
				  "((if \"%%I\"==\"DeviceNumber\" set A=%%J)&if \"%%J\"==\"\\BaseNamedObjects\\Global\\%s\" ArsenalImageMounterControl.exe /remove /device=!A!)\"", comm_device);
	CreateProcessA(NULL, path, NULL, NULL, FALSE, CREATE_NO_WINDOW, NULL, NULL, &si, &pi);
	WaitForSingleObject(pi.hProcess, INFINITE);
	return 0;
}

static int do_comm()
{
	HANDLE shm_request_event, shm_response_event;
	unsigned char *shm_view, *main_buf;
	IMDPROXY_IO_REQ *req_block;
	IMDPROXY_IO_RESP *resp_block;
	HANDLE hFileMap;
	ULARGE_INTEGER map_size;
	char objname[40], *index;
	STARTUPINFO si = {sizeof si};
	PROCESS_INFORMATION pi;
	DWORD wait_result;

	sprintf(comm_device, "%I64x", _rdtsc());
	index = objname + sprintf(objname, "Global\\%s", comm_device);

	map_size.QuadPart = buffer_size + IMDPROXY_HEADER_SIZE;

	if (!(hFileMap = CreateFileMappingA(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE | SEC_COMMIT, map_size.HighPart, map_size.LowPart, objname))) {
		disp_err("CreateFileMapping() failed.\nPlease check that you have administrative privileges.");
		return 2;
	}
	if (!(shm_view = MapViewOfFile(hFileMap, FILE_MAP_WRITE, 0, 0, 0))) {
		disp_err("MapViewOfFile() failed.");
		return 2;
	}

	main_buf = shm_view + IMDPROXY_HEADER_SIZE;
	req_block = (IMDPROXY_IO_REQ*)shm_view;
	resp_block = (IMDPROXY_IO_RESP*)shm_view;

	strcpy(index, "_Request");
	if (!(shm_request_event = CreateEventA(NULL, FALSE, FALSE, objname))) {
		disp_err("CreateEvent() failed.");
		return 2;
	}
	if (GetLastError() == ERROR_ALREADY_EXISTS) {
		disp_err("Synchronization objects already exist.");
		return 2;
	}
	strcpy(index, "_Response");
	if (!(shm_response_event = CreateEventA(NULL, FALSE, FALSE, objname))) {
		disp_err("CreateEvent() failed.");
		return 2;
	}

	if (use_imdisk) {
		if (wcslen(mount_point) > 32766) mount_point[32766] = 0;
		if (wcslen(format_param) > 100) format_param[100] = 0;
		swprintf(file_name, L"imdisk.exe -a -t proxy -m \"%s\" -o shm -f %S -S %d%S%s%S", mount_point, comm_device, sector_size, format_param[0] ? " -p \"" : "", format_param, format_param[0] ? "\"" : "");
	} else swprintf(file_name, L"ArsenalImageMounterControl.exe /add /filename=%S /sectorsize=%d /proxy=shm", comm_device, sector_size);
	CreateProcess(NULL, file_name, NULL, NULL, FALSE, verbose ? 0 : CREATE_NO_WINDOW, NULL, NULL, &si, &pi);
	FreeConsole();
	detach_console = TRUE;

	if (WaitForSingleObject(shm_request_event, use_imdisk ? 10000 : 20000) != WAIT_OBJECT_0) {
		disp_err("WaitForSingleObject() failed.");
		return 2;
	}
	if (req_block->request_code != IMDPROXY_REQ_INFO) {
		disp_err("Incorrect first request.");
		return 2;
	}

	out("Connection on object %s.\n", comm_device);

	memcpy(shm_view, &proxy_info, sizeof proxy_info);

	for (;;) {
		if (!SetEvent(shm_response_event)) {
			disp_err("SetEvent() failed.");
			return 1;
		}
		while ((wait_result = WaitForSingleObject(shm_request_event, timeout)) != WAIT_OBJECT_0) {
			if (wait_result == WAIT_TIMEOUT) {
				out("Inactivity period elapsed.\n");
				if (use_imdisk) {
					swprintf(file_name, L"imdisk.exe -D -m \"%s\"", mount_point);
					CreateProcess(NULL, file_name, NULL, NULL, FALSE, CREATE_NO_WINDOW, NULL, NULL, &si, &pi);
				} else
					CreateThread(NULL, 0, aim_dismount, NULL, 0, NULL);
			} else {
				disp_err("Error while flushing comm data.");
				return 1;
			}
		}
		if (req_block->request_code == IMDPROXY_REQ_READ)
			resp_block->errorno = physical_read(main_buf, req_block->length, offset + req_block->offset);
		else if (req_block->request_code == IMDPROXY_REQ_WRITE)
			resp_block->errorno = physical_write(main_buf, req_block->length, offset + req_block->offset);
		else {
			out("Connection closed.\n");
			SecureZeroMemory(main_buf, buffer_size);
			WaitForSingleObject(overlap_write.hEvent, INFINITE);
			SecureZeroMemory(write_buf, buffer_size);
			return 0;
		}
		resp_block->length = req_block->length;
	}
}

#pragma GCC optimize "Os"

__declspec(noinline) static void raw_physical_read(void *buf, int size, __int64 offset)
{
	DWORD bytes_read;
	__int64 *overlap_read_Offset = (__int64*)&overlap_read.Offset;

	*overlap_read_Offset = offset;
	ReadFile(h, buf, size, NULL, &overlap_read);
	if (verbose >= 2) out(" ? Read: %d bytes at offset %I64d\n", size, offset);
	GetOverlappedResult(h, &overlap_read, &bytes_read, TRUE);
	if (bytes_read != size)
		fprintf(stderr, "%%Read error: %u / %d bytes at offset %I64d.\n", (unsigned int)bytes_read, size, *overlap_read_Offset);
}

__declspec(noinline) static void raw_physical_write(void *buf, int size, __int64 offset)
{
	DWORD bytes_written;
	__int64 *overlap_write_Offset = (__int64*)&overlap_write.Offset;

	GetOverlappedResult(h, &overlap_write, &bytes_written, TRUE);
	if (bytes_written != expected_write)
		fprintf(stderr, "%%Write error: %u / %d bytes at offset %I64d.\n", (unsigned int)bytes_written, expected_write, *overlap_write_Offset);
	memcpy(write_buf, buf, size);
	*overlap_write_Offset = offset;
	WriteFile(h, write_buf, size, NULL, &overlap_write);
	expected_write = size;
	if (verbose >= 2) out(" ! Write: %d bytes at offset %I64d\n", size, offset);
}

static LRESULT __stdcall WndProc(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	HANDLE device;
	DWORD dw;
	DEV_BROADCAST_VOLUME dbv;

	if (Msg == WM_ENDSESSION || (Msg == WM_POWERBROADCAST && wParam == PBT_APMSUSPEND && suspend_close)) {
		ResetEvent(exit_event);
		if (use_imdisk) {
			device = (HANDLE)ImDiskOpenDeviceByMountPoint(mount_point, GENERIC_READ);
			FlushFileBuffers(device);
			if (!DeviceIoControl(device, FSCTL_DISMOUNT_VOLUME, NULL, 0, NULL, 0, &dw, NULL) || !DeviceIoControl(device, IOCTL_STORAGE_EJECT_MEDIA, NULL, 0, NULL, 0, &dw, NULL))
				ImDiskForceRemoveDevice(device, 0);
			CloseHandle(device);
			ImDiskRemoveMountPoint(mount_point);
			if (Msg != WM_ENDSESSION && mount_point[1] == ':' && !mount_point[2]) {
				SHChangeNotify(SHCNE_DRIVEREMOVED, SHCNF_PATH, mount_point, NULL);
				dbv.dbcv_size = sizeof dbv;
				dbv.dbcv_devicetype = DBT_DEVTYP_VOLUME;
				dbv.dbcv_unitmask = 1 << (mount_point[0] - 'A');
				dbv.dbcv_flags = 0;
				SendMessageTimeout(HWND_BROADCAST, WM_DEVICECHANGE, DBT_DEVICEREMOVECOMPLETE, (LPARAM)&dbv, SMTO_BLOCK | SMTO_ABORTIFHUNG, 100, NULL);
			}
		} else
			aim_dismount((LPVOID)(size_t)(Msg == WM_ENDSESSION));
		SignalObjectAndWait(exit_event, clean_ok, 2000, FALSE);
		return 0;
	} else
		return DefWindowProc(hWnd, Msg, wParam, lParam);
}

static DWORD __stdcall msg_window(LPVOID lpParam)
{
	MSG msg;
	WNDCLASSA wc = {0};

	wc.lpfnWndProc = WndProc;
	wc.lpszClassName = "X";
	RegisterClassA(&wc);
	CreateWindowA("X", NULL, 0, 0, 0, 0, 0, NULL, NULL, NULL, NULL);

	for (;;) {
		GetMessage(&msg, NULL, 0, 0);
		DispatchMessage(&msg);
	}
	return 0;
}

static void init_key(unsigned char key[6][32])
{
	long c = dec_cipher;

	n_cipher = 0;
	do {
		switch (c & 0xff) {
			case CIPHER_AES:
				aes_init(&keys[n_cipher].aes, key[n_cipher]);
				aes_init(&keys[n_cipher + 1].aes, key[n_cipher + 1]);
				break;
			case CIPHER_SERPENT:
				serpent_init(&keys[n_cipher].serpent, key[n_cipher]);
				serpent_init(&keys[n_cipher + 1].serpent, key[n_cipher + 1]);
				break;
			case CIPHER_TWOFISH:
				twofish_init(&keys[n_cipher].twofish, key[n_cipher]);
				twofish_init(&keys[n_cipher + 1].twofish, key[n_cipher + 1]);
		}
		n_cipher += 2;
	} while (c >>= 8);
}

static char *group_digit(unsigned long long n)
{
	int size = sprintf(global_str, "%I64u", n);

	while (size > 3) {
		size -= 3;
		memmove(global_str + size + 1, global_str + size, 24);
		global_str[size] = ' ';
	}
	return global_str;
}

__declspec(noinline) static void process_suffix(__int64 *size, TCHAR suffix)
{
	switch (suffix) {
		case 't': case 'T': *size <<= 10;
		case 'g': case 'G': *size <<= 10;
		case 'm': case 'M': *size <<= 10;
		case 'k': case 'K': *size <<= 10;
	}
}

static void format_size(__int64 size, int drive_number, __int64 offset, VOLUME_DISK_EXTENTS *vde)
{
	__int64 size_disp = size;
	char *suffix = "  KBMBGBTBPB";
	int i;

	while (size_disp > 9999) size_disp >>= 10, suffix += 2;
	sprintf(global_str, "%6I64d %s", size_disp, suffix);

	memset(&global_str[9], ' ', 14);
	if (offset >= 0) {
		for (i = 0; i < 26; i++)
			if (vde[i].Extents[0].DiskNumber == drive_number && vde[i].Extents[0].StartingOffset.QuadPart == offset) {
				global_str[11] = '(';
				global_str[12] = 'A' + i;
				global_str[13] = ':';
				global_str[14] = ')';
			}
		sprintf(&global_str[17], "[%I64d - %I64d]", offset, offset + size - 1);
	} else
		sprintf(&global_str[23], "[0 - %I64d]", size - 1);
}

static void display_drives()
{
	int drive_number, partition_number, ext_number;
	int entry_size, i;
	unsigned char *mbr;
	unsigned char partition_type;
	DISK_GEOMETRY_EX dge;
	DWORD n_bytes;
	__int64 drive_size, partition_size, offset_ext;
	VOLUME_DISK_EXTENTS vde[26];

	for (i = 0; i < 26; i++) {
		sprintf(global_str, "\\\\.\\%c:", 'A' + i);
		h = CreateFileA(global_str, 0, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (!DeviceIoControl(h, IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS, NULL, 0, &vde[i], sizeof(VOLUME_DISK_EXTENTS), &n_bytes, NULL))
			vde[i].Extents[0].DiskNumber = -1;
		CloseHandle(h);
	}

	mbr = VirtualAlloc(NULL, 34 * 4096, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE);
	for (drive_number = 0; ; drive_number++) {
		sprintf(global_str, "\\\\.\\PhysicalDrive%d", drive_number);
		if ((h = CreateFileA(global_str, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL)) == INVALID_HANDLE_VALUE) break;
		sector_size = 512;
		drive_size = 0;
		if (DeviceIoControl(h, IOCTL_DISK_GET_DRIVE_GEOMETRY_EX, NULL, 0, &dge, sizeof dge, &n_bytes, NULL)) {
			sector_size = dge.Geometry.BytesPerSector;
			drive_size = dge.DiskSize.QuadPart;
		}
		if (sector_size > 4096) continue;
		format_size(drive_size, drive_number, -1, vde);
		printf("Drive %d:%s\n", drive_number, global_str);
		raw_physical_read(mbr, 2 * 4096, 0);
		if (mbr[0x01BE + 4] == 0xEE && !memcmp(mbr + sector_size, "EFI PART\0\0\1", 12)) {
			entry_size = *(int*)(mbr + sector_size + 84);
			raw_physical_read(mbr, 34 * sector_size, 0);
			for (partition_number = 0; partition_number < 128; partition_number++) {
				i = ((partition_number >> 2) + 2) * sector_size + (partition_number & 3) * entry_size;
				if (!(*(__int64*)(mbr + i) | *(__int64*)(mbr + i + 8))) break;
				offset = *(__int64*)(mbr + i + 32) * sector_size;
				partition_size = (*(__int64*)(mbr + i + 40) + 1) * sector_size - offset;
				format_size(partition_size, drive_number, offset, vde);
				printf(" -partition %d:%s\n", partition_number + 1, global_str);
			}
		}
		else if (*(USHORT*)(mbr + 0x01FE) == 0xAA55 && ((mbr[0x01BE] | mbr[0x01CE] | mbr[0x01DE] | mbr[0x01EE]) & 0x7F) == 0) {
			for (partition_number = 1; partition_number <= 4; partition_number++) {
				partition_size = (__int64)(*(UINT*)(mbr + 0x01AE + partition_number * 16 + 12)) * sector_size;
				if (!partition_size) break;
				format_size(partition_size, drive_number, (__int64)(*(UINT*)(mbr + 0x01AE + partition_number * 16 + 8)) * sector_size, vde);
				printf(" -partition %d:%s\n", partition_number, global_str);
				partition_type = mbr[0x01AE + partition_number * 16 + 4];
				if (partition_type == 0x05 || partition_type == 0x0F) {
					offset_ext = offset = (__int64)(*(UINT*)(mbr + 0x01AE + partition_number * 16 + 8)) * sector_size;
					ext_number = 5;
					do {
						raw_physical_read(mbr, 4096, offset);
						partition_size = (__int64)(*(UINT*)(mbr + 0x01BE + 12)) * sector_size;
						format_size(partition_size, drive_number, offset + (__int64)(*(UINT*)(mbr + 0x01BE + 8)) * sector_size, vde);
						printf(" --partition %d:%s\n", ext_number++, global_str);
						offset = offset_ext + (__int64)(*(UINT*)(mbr + 0x01CE + 8)) * sector_size;
					} while (offset != offset_ext);
				}
			}
		}
		CloseHandle(h);
		puts("");
	}
}

static void benchmark(unsigned long cipher)
{
	HCRYPTPROV hProv;
	DWORD init_time, enc_time, dec_time, n_enc, n_dec;
	int i, j;
	unsigned long cipher_list[] = {1, 2, 3, 0x0301, 0x0102, 0x0302, 0x030102};
	char cipher_str[32];

	thread_buffer_read = thread_buffer_write = VirtualAlloc(NULL, DEF_BUFFER_SIZE, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE);
	request_size_read = request_size_write = DEF_BUFFER_SIZE;
	thread_offset_read = thread_offset_write = 0;
	if (!CryptAcquireContext(&hProv, NULL, NULL, PROV_RSA_FULL, 0))
		CryptAcquireContext(&hProv, NULL, NULL, PROV_RSA_FULL, CRYPT_NEWKEYSET);
	CryptGenRandom(hProv, DEF_BUFFER_SIZE, thread_buffer_read);
	puts("The following tests are made in RAM in XTS mode.\n\nCIPHER\t\t       ENCRYPT\t DECRYPT   AVERAGE (MB/s)\n");
	init_time = GetTickCount();
	do {} while (GetTickCount() - init_time < 50); // ensure that CPU is at correct frequency
	for (j = 0; j < _countof(cipher_list); j++) {
		dec_cipher = cipher ? cipher : cipher_list[j];
		enc_cipher = _byteswap_ulong(dec_cipher);
		do enc_cipher >>= 8; while (!(enc_cipher & 0xff));
		init_key((void*)thread_buffer_read);
		n_enc = n_dec = 0;
		init_time = GetTickCount();
		do {
			i = 0; do SetEvent(encrypt_event[i]); while (++i < si.dwNumberOfProcessors);
			WaitForMultipleObjects(si.dwNumberOfProcessors, encrypt_done, TRUE, INFINITE);
			enc_time = GetTickCount();
			n_enc++;
		} while (enc_time - init_time < 500);
		do {
			i = 0; do SetEvent(decrypt_event[i]); while (++i < si.dwNumberOfProcessors);
			WaitForMultipleObjects(si.dwNumberOfProcessors, decrypt_done, TRUE, INFINITE);
			dec_time = GetTickCount();
			n_dec++;
		} while (dec_time - enc_time < 500);
		sprintf(cipher_str, "%s%s%s%s%s", cipher_name[dec_cipher & 0xff], dec_cipher & 0xff00 ? "-" : "", cipher_name[(dec_cipher >> 8) & 0xff], dec_cipher & 0xff0000 ? "-" : "", cipher_name[dec_cipher >> 16]);
		printf("%-23s%7u%10u%10u\n", cipher_str,
			   (UINT)((__int64)n_enc * DEF_BUFFER_SIZE * 1000 / (enc_time - init_time) >> 20),
			   (UINT)((__int64)n_dec * DEF_BUFFER_SIZE * 1000 / (dec_time - enc_time) >> 20),
			   (UINT)((__int64)(n_enc + n_dec) * DEF_BUFFER_SIZE * 1000 / (dec_time - init_time) >> 20));
		if (cipher) break;
	}
}

int wmain(int argc, TCHAR *argv[])
{
	int drive_number = -1;
	int partition_number = 1;
	BOOL force_partition = FALSE;
	int entry_size;
	unsigned char *mbr;
	__int64 offset_ext;
	BOOL new_volume = FALSE, change_pass = FALSE, bench = FALSE;
	int n_hash = 0;
	__int64 current_size, new_file_size = 0;
	__int64 max_volume_size, size_new_volume = 0;
	DISK_GEOMETRY_EX dge;
	DWORD n_bytes;
	TCHAR suffix, *prompt_buf, *p1 = NULL, *p2 = NULL, *p3 = NULL;
	char path[MAX_PATH];
	int i, j, retval = 0;
	HEADER_STRUCT *header;
	union {unsigned char ch[512 + 64]; HEADER_STRUCT header; __m128i i128;} tmp_header;
	unsigned char *buf;
	unsigned char header_key[6][32];
	HANDLE h_backup = NULL, h_find;
	HANDLE thread_handle[2];
	int CPUInfo[4];
	BOOL mount = FALSE;
	WIN32_FIND_DATA FindFileData;
	unsigned long imdisk_version = 0;

	if (argc < 2) {
		PathStripPath(argv[0]);
		PathRemoveExtension(argv[0]);
		i = min(wcslen(argv[0]), 29);
		memset(global_str, ' ', i);
		global_str[i] = 0;
		printf("version 1.6.1\n\n"
			"-= Syntax =-\n"
			"%S [mount_point] -f file | -d drivenumber[,[partition]]\n"
			"%s [-c [size]] [-p partition] [-o offset] [-cp] [-r]\n"
			"%s [-ad time] [-ads] [-t threads] [-b backup_file]\n"
			"%s [-fp \"arguments\"] [-pp ...] [-v verbose] [-aim]\n"
			"%S -l\n"
			"%S -bm [cipher[,cipher[,cipher]]] [-t threads]\n\n"
			" mount_point\n"
			"\tSpecifies a drive letter (followed by a colon) or an empty directory\n"
			"\tlocated on a NTFS volume. If omitted, a drive letter is selected\n"
			"\tautomatically.\n\n"
			" -f\tImage file of the encrypted volume.\n\n"
			" -c\tCreate a new encrypted volume. If a size and a file name are specified,\n"
			"\tan image file of this size will be created. Use k, m, g or t as suffix\n"
			"\tto specify KB, MB, GB or TB.\n\n"
			" -p\tNumber of partition (default: 1). Use 0 to ignore the partition table\n"
			"\t(if any) and to use the whole disk or image.\n\n"
			" -d\tNumber of physical drive where the encrypted volume is located.\n"
			"\tIf a comma is added without partition number, the latter is assumed to\n"
			"\tbe 0 instead of 1.\n\n"
			" -o\tOffset to the encrypted volume inside the file/drive/partition.\n"
			"\tUse k, m, g, or t as suffix to specify KB, MB, GB or TB.\n\n"
			" -cp\tChange password.\n\n"
			" -r\tMount a volume as read-only.\n\n"
			" -ad\tAuto-dismount after a period of inactivity, in minutes, from 1 to 71582\n\n"
			" -ads\tAuto-dismount when system enters a suspended state.\n\n"
			" -t\tNumber of threads used for decryption and encryption, from 1 to 32\n"
			"\t(the default value is the number of logical processors).\n\n"
			" -b\tIf backup_file does not exist, it will be created with the content of\n"
			"\tthe volume header, encrypted with a new salt.\n"
			"\tIf backup_file exists, the volume header will be overwritten with the\n"
			"\tfirst 512 bytes of backup_file.\n\n"
			" -fp\tFormat parameters. \"arguments\" is a quoted string passed to the format\n"
			"\tcommand. Ignored with Arsenal Image Mounter.\n\n"
			" -pp\tParameters entered at the prompts can be put here. Intended to\n"
			"\tsemi-automated volume creation (3 parameters in same format and order)\n"
			"\tor password change (1 parameter). -cp must be put before this switch.\n\n"
			" -v\tVerbosity level, from 0 to 2 (default: 1).\n\n"
			" -aim\tForce the use of Arsenal Image Mounter, in case of the ImDisk Virtual\n"
			"\tDisk Driver, which is used by default, is also installed.\n"
			"\tIf Arsenal Image Mounter is used, mount_point is ignored.\n\n"
			" -l\tDisplay a list of current physical drives and partitions, with location\n"
			"\tof first and last bytes.\n\n"
			" -bm\tBenchmark of available ciphers.\n"
			"\tA cipher number can be specified, with 1=AES, 2=Serpent and 3=Twofish.\n"
			"\tUse 2 or 3 numbers, separated by commas, for cascaded ciphers.\n\n\n"
			"-= Password editor usage =-\n"
			" Esc:\tQuit without volume mounting.\n\n"
			" Tab:\tShow/hide password.\n\n"
			" Enter:\tUsed with an empty password, it disables low level access to the\n"
			"\tkeyboard. This improves compatibility with some Input Method Editors\n"
			"\tbut removes protection against keyloggers.\n",
			argv[0], global_str, global_str, global_str, argv[0], argv[0]);
		return 0;
	}

	GetSystemInfo(&si);
	__cpuid(CPUInfo, 1);
	if (!(CPUInfo[3] & 0x4000000)) {
		disp_err("SSE2 instructions required.");
		return 1;
	}

	if (!_wcsicmp(argv[1], L"-l")) {
		display_drives();
		return 0;
	}

	argc--;
	file_name[0] = 0;
	do {
		argv++;
		if (!_wcsicmp(argv[0], L"-bm")) {
			bench = TRUE;
			i = j = 0;
			if (argc > 1 && swscanf(argv[1], L"%d,%d,%d", &dec_cipher, &i, &j)) {
				if (dec_cipher < 1 || dec_cipher > 3) dec_cipher = 1;
				if (i < 0 || i > 3) i = 0;
				if (j < 0 || j > 3) j = 0;
				dec_cipher += (i << 8) + (j << 16);
				argc--;
				argv++;
			}
			continue;
		}
		if (!_wcsicmp(argv[0], L"-f")) {
			if (--argc <= 0) goto syntax_error;
			argv++;
			if (wcslen(argv[0]) < MAX_PATH)
				wcscpy(file_name, argv[0]);
			else {
				_snwprintf(file_name, _countof(file_name) - 1, L"\\\\?\\%s", argv[0]);
				file_name[_countof(file_name) - 1] = 0;
			}
			continue;
		}
		if (!_wcsicmp(argv[0], L"-c")) {
			new_volume = TRUE;
			if (argc > 1) {
				suffix = 0;
				if (swscanf(argv[1], L"%I64d%c", &new_file_size, &suffix)) {
					process_suffix(&new_file_size, suffix);
					if (new_file_size < 24 * 1024) {
						disp_err("New file size must be at least 24 KB.");
						return 1;
					}
					argc--;
					argv++;
				}
			}
			continue;
		}
		if (!_wcsicmp(argv[0], L"-p")) {
			if (--argc <= 0) goto syntax_error;
			argv++;
			swscanf(argv[0], L"%u", &partition_number);
			force_partition = TRUE;
			continue;
		}
		if (!_wcsicmp(argv[0], L"-d")) {
			if (--argc <= 0) goto syntax_error;
			argv++;
			switch (swscanf(argv[0], L"%u%c%u", &drive_number, &suffix, &partition_number)) {
				case 2: partition_number = 0;
				case 3: force_partition = TRUE;
			}
			continue;
		}
		if (!_wcsicmp(argv[0], L"-o")) {
			if (--argc <= 0) goto syntax_error;
			argv++;
			suffix = 0;
			swscanf(argv[0], L"%I64d%c", &extra_offset, &suffix);
			process_suffix(&extra_offset, suffix);
			continue;
		}
		if (!_wcsicmp(argv[0], L"-cp")) {
			change_pass = TRUE;
			continue;
		}
		if (!_wcsicmp(argv[0], L"-r")) {
			proxy_info.flags = IMDPROXY_FLAG_RO;
			continue;
		}
		if (!_wcsicmp(argv[0], L"-ad")) {
			if (--argc <= 0) goto syntax_error;
			argv++;
			swscanf(argv[0], L"%u", &timeout);
			if (timeout < 1) timeout = 1;
			if (timeout > 71582) timeout = 71582;
			timeout *= 60000;
			continue;
		}
		if (!_wcsicmp(argv[0], L"-ads")) {
			suspend_close = TRUE;
			continue;
		}
		if (!_wcsicmp(argv[0], L"-t")) {
			if (--argc <= 0) goto syntax_error;
			argv++;
			swscanf(argv[0], L"%u", &si.dwNumberOfProcessors);
			continue;
		}
		if (!_wcsicmp(argv[0], L"-b")) {
			if (--argc <= 0) goto syntax_error;
			argv++;
			backup_file = argv[0];
			continue;
		}
		if (!_wcsicmp(argv[0], L"-fp")) {
			if (--argc <= 0) goto syntax_error;
			argv++;
			format_param = argv[0];
			continue;
		}
		if (!_wcsicmp(argv[0], L"-pp")) {
			if (argc < 2) goto syntax_error;
			use_prompt = FALSE;
			p1 = argv[1];
			argc--;
			argv++;
			if (!change_pass) {
				if (argc < 3) goto syntax_error;
				p2 = argv[1];
				p3 = argv[2];
				argc -= 2;
				argv += 2;
			}
			continue;
		}
		if (!_wcsicmp(argv[0], L"-v")) {
			if (--argc <= 0) goto syntax_error;
			argv++;
			verbose = *argv[0] - '0';
			continue;
		}
		if (!_wcsicmp(argv[0], L"-aim")) {
			force_aim = TRUE;
			continue;
		}
		if (argv[0][1] == ':' && argv[0][2] == 0) {
			if (argv[0][0] < 'A' || (argv[0][0] > 'Z' && argv[0][0] < 'a') || argv[0][0] > 'z') {
syntax_error:
				disp_err("Syntax error.");
				return 1;
			}
			if (PathFileExists(argv[0]))
				fwprintf(stderr, L"Warning: %s already in use.\n", argv[0]);
		} else {
			FindClose(h_find = FindFirstFile(argv[0], &FindFileData));
			if (h_find == INVALID_HANDLE_VALUE || ((FindFileData.dwFileAttributes & (FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_REPARSE_POINT)) == FILE_ATTRIBUTE_DIRECTORY && !PathIsDirectoryEmpty(argv[0]))) {
				fwprintf(stderr, L"%s is not an empty directory.\n", argv[0]);
				return 1;
			}
		}
		mount_point = argv[0];
	} while (--argc > 0);

	if (CPUInfo[2] & 0x2000000) {
		out("AES instructions detected.\n");
		aes_encrypt = aes_encrypt_hw;
		aes_decrypt = aes_decrypt_hw;
	} else {
		aes_encrypt = aes_encrypt_sw;
		aes_decrypt = aes_decrypt_sw;
	}

	// Create threads
	if (si.dwNumberOfProcessors < 1) si.dwNumberOfProcessors = 1;
	if (si.dwNumberOfProcessors > 32) si.dwNumberOfProcessors = 32;
	out("Uses %u thread%s for encryption/decryption.\n", si.dwNumberOfProcessors, si.dwNumberOfProcessors > 1 ? "s" : "");
	i = 0; do {
		if (!(encrypt_event[i] = CreateEvent(NULL, FALSE, FALSE, NULL)) || !(decrypt_event[i] = CreateEvent(NULL, FALSE, FALSE, NULL)) ||
			!(encrypt_done[i] = CreateEvent(NULL, FALSE, FALSE, NULL)) || !(decrypt_done[i] = CreateEvent(NULL, FALSE, FALSE, NULL))) {
			disp_err("CreateEvent() failed.");
			return 1;
		}
		if (!CreateThread(NULL, 0, encrypt_thread, (LPVOID)(size_t)i, 0, NULL) || !CreateThread(NULL, 0, decrypt_thread, (LPVOID)(size_t)i, 0, NULL)) {
			disp_err("CreateThread() failed.");
			return 1;
		}
	} while (++i < si.dwNumberOfProcessors);

	overlap_read.hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
	overlap_write.hEvent = CreateEvent(NULL, FALSE, TRUE, NULL);

	if (bench) {
		benchmark(dec_cipher);
		return 0;
	}

	strcpy(path, "imdisk.exe");
	use_imdisk = PathFileExistsA(path) || PathFindOnPathA(path, NULL);
	if (!use_imdisk || force_aim) {
		strcpy(path, "ArsenalImageMounterControl.exe");
		if (PathFileExistsA(path) || PathFindOnPathA(path, NULL)) {
			use_imdisk = FALSE;
			buffer_size = AIM_BUFFER_SIZE;
		} else {
			disp_err(force_aim ? "Cannot find ArsenalImageMounterControl.exe." : "Can find neither imdisk.exe nor ArsenalImageMounterControl.exe.");
			return 1;
		}
	}
	if (use_imdisk) {
		if (!(h_cpl = LoadLibraryA("imdisk.cpl"))) {
			disp_err("Cannot find imdisk.cpl.");
			return 1;
		}
		ImDiskOpenDeviceByMountPoint = GetProcAddress(h_cpl, "ImDiskOpenDeviceByMountPoint");
		ImDiskForceRemoveDevice = GetProcAddress(h_cpl, "ImDiskForceRemoveDevice");
		ImDiskRemoveMountPoint = GetProcAddress(h_cpl, "ImDiskRemoveMountPoint");
		GetProcAddress(h_cpl, "ImDiskGetVersion")(&imdisk_version, (PULONG)NULL);
		if (imdisk_version < 0x0176) {
			disp_err("ImDisk 1.7.6 or later required.");
			return 1;
		}
		if (!mount_point) {
			if ((drive_letter[0] = _bit_scan_forward(~(GetLogicalDrives() >> 3)) + 'D') > 'Z') {
				disp_err("Cannot find free drive letter.");
				return 1;
			}
			drive_letter[1] = ':';
			drive_letter[2] = 0;
			mount_point = drive_letter;
		}
	}

	if (new_file_size) {
		if (proxy_info.flags) {
			disp_err("Cannot create file in read-only mode.");
			return 1;
		}
		if ((h = CreateFile(file_name, 0, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_BACKUP_SEMANTICS, NULL)) != INVALID_HANDLE_VALUE) {
			CloseHandle(h);
			disp_err("Error: a file or folder of this name already exists.");
			return 1;
		}
		proxy_info.file_size = new_file_size;
	} else {
		// Opening file or drive
		if (drive_number == -1) {
			if ((h = CreateFile(file_name, proxy_info.flags ? GENERIC_READ : GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_EXISTING,
								FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED | FILE_FLAG_WRITE_THROUGH, NULL)) == INVALID_HANDLE_VALUE) {
				fwprintf(stderr, L"Cannot open %s.\n", file_name);
				return 1;
			}
			GetFileTime(h, NULL, &access_time, &write_time);
			GetFileSizeEx(h, (LARGE_INTEGER*)&proxy_info.file_size);
		} else {
			swprintf(file_name, L"\\\\.\\PhysicalDrive%d", drive_number);
			if ((h = CreateFile(file_name, proxy_info.flags ? GENERIC_READ : GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING,
								FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED | FILE_FLAG_NO_BUFFERING | FILE_FLAG_WRITE_THROUGH, NULL)) == INVALID_HANDLE_VALUE) {
				fprintf(stderr, "Cannot open drive %d.\n", drive_number);
				return 1;
			}
			// Getting sector size
			if (DeviceIoControl(h, IOCTL_DISK_GET_DRIVE_GEOMETRY_EX, NULL, 0, &dge, sizeof dge, &n_bytes, NULL)) {
				sector_size = dge.Geometry.BytesPerSector;
				proxy_info.file_size = dge.DiskSize.QuadPart;
			}
			if (sector_size < 512 || sector_size > 4096) {
				fprintf(stderr, "\nSector size must be at least 512-bytes and at most 4096-bytes\n(detected: %d)\n", sector_size);
				return 1;
			}
		}
	}

	if (!proxy_info.file_size)
		disp_err("Cannot determine size of image/drive.");
	current_size = proxy_info.file_size;

	write_buf = VirtualAlloc(NULL, buffer_size, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE);
	buf = VirtualAlloc(NULL, 4096 * 5 + DEF_BUFFER_SIZE, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE);
	header = (HEADER_STRUCT*)(buf + 4096 * 2);
	password = (TCHAR*)(buf + 4096 * 4);
	prompt_buf = (TCHAR*)(buf + 4096 * 5);

	// Getting offset to the partition
	if (!new_file_size && partition_number >= 1 && partition_number <= 128 && current_size >= 2 * sector_size) {
		mbr = VirtualAlloc(NULL, 34 * sector_size, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE);
		raw_physical_read(mbr, 2 * 4096, 0);

		if (mbr[0x01BE + 4] == 0xEE && !memcmp(mbr + sector_size, "EFI PART\0\0\1", 12)) {
			out("GPT detected\n");

			entry_size = *(int*)(mbr + sector_size + 84);
			raw_physical_read(mbr, 34 * sector_size, 0);
			i = (((partition_number - 1) >> 2) + 2) * sector_size + ((partition_number - 1) & 3) * entry_size;
			if (*(__int64*)(mbr + i) | *(__int64*)(mbr + i + 8)) {
				offset = *(__int64*)(mbr + i + 32) * sector_size;
				proxy_info.file_size = (*(__int64*)(mbr + i + 40) + 1) * sector_size - offset;
			} else if (force_partition) {
				fprintf(stderr, "Partition %i not found.\n", partition_number);
				goto clean;
			}
		}
		else {
			if (*(USHORT*)(mbr + 0x01FE) == 0xAA55 && ((mbr[0x01BE] | mbr[0x01CE] | mbr[0x01DE] | mbr[0x01EE]) & 0x7F) == 0) {
				out("MBR detected.\n");

				if (partition_number >= 5) {
					for (i = 0x01BE; i < 0x01FE; i += 16)
						if (mbr[i + 4] == 0x05 || mbr[i + 4] == 0x0F) {
							offset_ext = offset = (__int64)(*(UINT*)(mbr + i + 8)) * sector_size;
							raw_physical_read(mbr, 4096, offset);
							for (j = 5; j < partition_number; j++) {
								offset = offset_ext + (__int64)(*(UINT*)(mbr + 0x01CE + 8)) * sector_size;
								if (offset == offset_ext) {
									fprintf(stderr, "Partition %i not found.\n", partition_number);
									goto clean;
								}
								raw_physical_read(mbr, 4096, offset);
							}
							offset += (__int64)(*(UINT*)(mbr + 0x01BE + 8)) * sector_size;
							proxy_info.file_size = (__int64)(*(UINT*)(mbr + 0x01BE + 12)) * sector_size;
							break;
						}
					if (i == 0x01FE) {
						disp_err("No extended partition found.");
						goto clean;
					}
				} else {
					offset = (__int64)(*(UINT*)(mbr + 0x01AE + partition_number * 16 + 8)) * sector_size;
					proxy_info.file_size = (__int64)(*(UINT*)(mbr + 0x01AE + partition_number * 16 + 12)) * sector_size;
				}

				if (proxy_info.file_size)
					out("Partition %i used.\n", partition_number);
				else {
					if (force_partition) {
						fprintf(stderr, "Partition %i not found.\n", partition_number);
						goto clean;
					} else {
						offset = 0;
						proxy_info.file_size = current_size;
					}
				}
			} else
				out("No MBR/GPT detected. Uses entire image.\n");
		}
		VirtualFree(mbr, 0, MEM_RELEASE);
	}

	proxy_info.req_alignment = DEF_REQUIRED_ALIGNMENT;
	out("Total size: %s bytes.\n", group_digit(current_size));
	out("%s bytes used from offset ", group_digit(proxy_info.file_size));
	out("%s.\n", group_digit(offset));

	offset += extra_offset;
	if (proxy_info.file_size < 20992) {
		disp_err("Too small volume.");
		goto clean;
	}
	if (proxy_info.file_size - extra_offset < 20992) {
		disp_err("Offset to the encrypted volume too high.");
		goto clean;
	}

	if (!VirtualLock(buf, 4096 * 5) |
		!VirtualLock(keys, sizeof keys) |
		!VirtualLock(header_key_whirlpool, sizeof header_key_whirlpool) |
		!VirtualLock(header_key_sha3, sizeof header_key_sha3) |
		!VirtualLock(&tmp_header, sizeof tmp_header))
		disp_err("Warning: cannot lock data into physical memory.");

	whirlpool_ini();

	if (new_volume) {
		tmp_header.header.volume_offset = (extra_offset + 512 + 4095) & ~4095;
		max_volume_size = proxy_info.file_size - tmp_header.header.volume_offset;
		suffix = 0;
		if (use_prompt) {
			printf("\n-Enter size of the encrypted volume (20 KB - %I64u KB).\n Use k, m, g, or t as suffix, 0 for the maximum available.\n", max_volume_size >> 10);
			fgetws(p1 = prompt_buf, DEF_BUFFER_SIZE / sizeof(TCHAR), stdin);
		}
		swscanf(p1, L"%I64d%c", &size_new_volume, &suffix);
		process_suffix(&size_new_volume, suffix);
		if (!size_new_volume) size_new_volume = max_volume_size;
		if (size_new_volume < 20 * 1024) size_new_volume = 20 * 1024;
		if (size_new_volume > max_volume_size) size_new_volume = max_volume_size;
		tmp_header.header.volume_size = size_new_volume;

		i = j = 0;
		if (use_prompt) {
			puts("\n-Select encryption algorithm.\n Use commas to combine up to 3 ciphers\n (e.g. 2,1 decrypts first with Serpent, then with AES)\n1: AES\n2: Serpent\n3: Twofish");
			fgetws(p2 = prompt_buf, DEF_BUFFER_SIZE / sizeof(TCHAR), stdin);
		}
		swscanf(p2, L"%d,%d,%d", &dec_cipher, &i, &j);
		if (dec_cipher < 1 || dec_cipher > 3 || (unsigned int)i > 3 || (unsigned int)j > 3) {
			printf("Unrecognized value. %s selected by default.\n", "AES");
			dec_cipher = 1;
		} else
			dec_cipher += (i << 8) + (j << 16);
		enc_cipher = _byteswap_ulong(dec_cipher);
		do enc_cipher >>= 8; while (!(enc_cipher & 0xff));

		if (use_prompt) {
			puts("\n-Select password hash algorithm.\n1: Whirlpool\n2: SHA-3-512");
			fgetws(p3 = prompt_buf, DEF_BUFFER_SIZE / sizeof(TCHAR), stdin);
		}
		swscanf(p3, L"%d", &n_hash);
		if (n_hash < 1 || n_hash > 2) {
			printf("Unrecognized value. %s selected by default.\n", "Whirlpool");
			n_hash = 1;
		}

		for (;;) {
			puts("\nEnter password:");
			if (password_edit(password)) goto clean;
			wcscpy((TCHAR*)buf, password);
			puts("Confirm password:");
			if (password_edit(password)) goto clean;
			if (!wcscmp(password, (TCHAR*)buf)) break;
			puts("Passwords are different. Please retry.");
		}

		if (new_file_size) {
			// Creating file
			if ((h = CreateFile(file_name, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_NEW, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED | FILE_FLAG_WRITE_THROUGH, NULL)) == INVALID_HANDLE_VALUE) {
				disp_err("Error: file cannot be created.");
				goto clean;
			}
			GetFileTime(h, NULL, &access_time, &write_time);
			SetFilePointerEx(h, (LARGE_INTEGER)new_file_size, NULL, FILE_BEGIN);
			if (!SetEndOfFile(h)) {
				disp_err("Cannot create a file of the specified size.");
				CloseHandle(h);
				DeleteFile(file_name);
				goto clean;
			}
			SetFilePointerEx(h, (LARGE_INTEGER)0LL, NULL, FILE_BEGIN);
		}

		out("Creating master keys...\n");
		if (rand_gen(buf, 512 + 4096, n_hash - 1)) goto clean;
		memcpy(tmp_header.header.salt, buf, 64);
		memcpy(tmp_header.header.master_key, buf + 64, sizeof header->master_key);
		whirlpool((unsigned char*)tmp_header.header.master_key, sizeof header->master_key, tmp_header.header.hash, 1, 1);

		out("Hashing password...\n");
		if (scrypt(n_hash - 1, password, wcslen(password) * sizeof(TCHAR), buf, 64, 32768, 16, 16, header_key, sizeof header_key)) goto clean;
		tmp_header.header.version = 1;

		raw_physical_read(header, 2 * 4096, offset & ~4095);
		init_key(header_key);
		XTS_EncryptSector(enc_cipher, offset, tmp_header.ch + 64);
		i = offset % 4096;
		memcpy((char*)header + i, tmp_header.ch, 512);
		memcpy((char*)header + i + 512, buf + 512, 4095 - (i + 511) % 4096);
		raw_physical_write(header, 2 * 4096, offset & ~4095);
		puts("Volume created successfully.");
		XTS_DecryptSector(dec_cipher, offset, tmp_header.ch + 64);
	}
	else
	{
		raw_physical_read(buf, 2 * 4096, offset & ~4095);
		memcpy(header, buf + (offset & 4095), 512);

		if (backup_file) {
			if (wcslen(backup_file) >= MAX_PATH) {
				_snwprintf(file_name, _countof(file_name) - 1, L"\\\\?\\%s", backup_file);
				file_name[_countof(file_name) - 1] = 0;
				backup_file = file_name;
			}
			if ((h_backup = CreateFile(backup_file, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL)) == INVALID_HANDLE_VALUE) {
				fwprintf(stderr, L"Cannot open %s.\n", backup_file);
				goto clean;
			}
			retval = GetLastError();
			if (retval == ERROR_ALREADY_EXISTS)
				if (!ReadFile(h_backup, (char*)header, 512, &n_bytes, NULL) || n_bytes != 512) {
					disp_err("Error while reading backup file.");
					goto clean;
				}
		}

enter_password:
		dec_cipher = n_hash = 0;
		puts(change_pass ? "\nEnter current password:" : "\nEnter password:");
		if (password_edit(password)) goto clean;

		printf("Checking password...");
		pass_size = wcslen(password) * sizeof(TCHAR);
		memcpy(salt_whirlpool, header->salt, 64);
		memcpy(salt_sha3, header->salt, 64);
		thread_handle[PRF_WHIRLPOOL] = CreateThread(NULL, 0, derive_pass_whirlpool, NULL, 0, NULL);
		thread_handle[PRF_SHA3] = CreateThread(NULL, 0, derive_pass_sha3, NULL, 0, NULL);
		if (WaitForMultipleObjects(_countof(thread_handle), thread_handle, TRUE, INFINITE) == WAIT_FAILED) {
			disp_err("\nWaitForMultipleObjects() failed.");
			goto clean;
		}
		do {
			if (dec_cipher == 0x030303) {
				if (++n_hash > 1) {
					puts(" Error. Please retry.");
					goto enter_password;
				}
				dec_cipher = 0;
			}
			dec_cipher++;
			if ((dec_cipher & 0xff) == 4) {
				dec_cipher += 0x100 - 3;
				if ((dec_cipher & 0xff00) == 0x400) dec_cipher += 0x10000 - 0x300;
			}
			init_key(n_hash ? header_key_sha3 : header_key_whirlpool);
			memcpy(tmp_header.ch, header, 512);
			XTS_DecryptSector(dec_cipher, offset, tmp_header.ch + 64);
			whirlpool((unsigned char*)tmp_header.header.master_key, sizeof header->master_key, buf, 1, 1);
		} while (memcmp(buf, tmp_header.header.hash, WHIRLPOOL_DIGESTSIZE));

		puts(" OK.");
		out("Password hashed with scrypt/%s.\nCipher is %s%s%s%s%s.\n", n_hash ? "SHA-3-512" : "Whirlpool",
			cipher_name[dec_cipher & 0xff], dec_cipher & 0xff00 ? "-" : "", cipher_name[(dec_cipher >> 8) & 0xff], dec_cipher & 0xff0000 ? "-" : "", cipher_name[dec_cipher >> 16]);

		enc_cipher = _byteswap_ulong(dec_cipher);
		do enc_cipher >>= 8; while (!(enc_cipher & 0xff));

		if (backup_file) {
			if (!retval) {
				out("Encrypting header with new salt...\n");
				memcpy(header, tmp_header.ch, 512);
				if (rand_gen(buf, 256, n_hash - 1)) goto clean;
				memcpy(header->salt, buf, 64);
				if (scrypt(n_hash - 1, password, pass_size, buf, 64, 32768, 16, 16, header_key, sizeof header_key)) goto clean;
				init_key(header_key);
				XTS_EncryptSector(enc_cipher, offset, (unsigned char*)header + 64);
				if (!WriteFile(h_backup, (char*)header, 512, &n_bytes, NULL) || n_bytes != 512) {
					disp_err("Error while saving volume header.");
					goto clean;
				}
				out("Volume header saved successfully.\n");
			}
			if (retval == ERROR_ALREADY_EXISTS) {
				memcpy((char*)buf + (offset & 4095), header, 512);
				raw_physical_write(buf, 2 * 4096, offset & ~4095);
				out("Volume header written successfully.\n");
			}
			CloseHandle(h_backup);
		}

		if (change_pass) {
			for (;;) {
				puts("\nEnter new password:");
				if (password_edit(password)) goto clean;
				wcscpy((TCHAR*)buf, password);
				puts("Confirm new password:");
				if (password_edit(password)) goto clean;
				if (!wcscmp(password, (TCHAR*)buf)) break;
				puts("Passwords are different. Please retry.");
			}
			if (use_prompt) {
				puts("\n-Select password hash algorithm:\n1: Whirlpool\n2: SHA-3-512");
				fgetws(p1 = prompt_buf, DEF_BUFFER_SIZE / sizeof(TCHAR) - 2 * 4096, stdin);
			}
			swscanf(p1, L"%d", &n_hash);
			if (n_hash < 1 || n_hash > 2) {
				printf("Unrecognized input. %s selected by default.\n", "Whirlpool");
				n_hash = 1;
			}
			out("Hashing new password...\n");
			if (rand_gen(buf, 256, n_hash - 1)) goto clean;
			memcpy(tmp_header.header.salt, buf, 64);
			if (scrypt(n_hash - 1, password, wcslen(password) * sizeof(TCHAR), buf, 64, 32768, 16, 16, header_key, sizeof header_key)) goto clean;
			init_key(header_key);
			XTS_EncryptSector(enc_cipher, offset, tmp_header.ch + 64);
			raw_physical_read(header, 2 * 4096, offset & ~4095);
			memcpy((char*)header + offset % 4096, tmp_header.ch, 512);
			raw_physical_write(header, 2 * 4096, offset & ~4095);
			puts("Password changed.");
			XTS_DecryptSector(dec_cipher, offset, tmp_header.ch + 64);
		}
	}

	init_key(tmp_header.header.master_key);
	offset += tmp_header.header.volume_offset - extra_offset;
	proxy_info.file_size = tmp_header.header.volume_size;
	out("Encrypted volume of %s bytes at offset ", group_digit(proxy_info.file_size));
	out("%s.\n\n", group_digit(offset));
	mount = TRUE;

clean:
	SecureZeroMemory(buf, 4096 * 5);
	VirtualFree(buf, 0, MEM_RELEASE);
	SecureZeroMemory(header_key_whirlpool, sizeof header_key_whirlpool);
	SecureZeroMemory(header_key_sha3, sizeof header_key_sha3);
	SecureZeroMemory(&tmp_header, sizeof tmp_header);
	SecureZeroMemory(header_key, sizeof header_key);

	exit_event = CreateEvent(NULL, FALSE, TRUE, NULL);
	clean_ok = CreateEvent(NULL, FALSE, FALSE, NULL);
	if (mount) {
		SetProcessShutdownParameters(0x100, 0);
		CreateThread(NULL, 0, msg_window, NULL, 0, NULL);
		retval = do_comm();
	}

	SecureZeroMemory(keys, sizeof keys);
	SetFileTime(h, NULL, &access_time, &write_time);
	CloseHandle(h);
	SignalObjectAndWait(clean_ok, exit_event, 5000, FALSE);

	return retval | !mount;
}

#ifdef MINGW_NOSTDLIB
void mainCRTStartup()
{
	int argc;
	LPWSTR *argv = CommandLineToArgvW(GetCommandLine(), &argc);
	ExitProcess(wmain(argc, argv));
}
#endif
