#include <windows.h>
#include <intrin.h>
#include "aes_hw.h"

static void key_256_assist_1(__m128i *temp1, __m128i temp2)
{
	__m128i temp4;

	temp2 = _mm_shuffle_epi32(temp2, 0xff);
	temp4 = _mm_slli_si128(*temp1, 4);
	*temp1 ^= temp4;
	temp4 = _mm_slli_si128(temp4, 4);
	*temp1 ^= temp4 ^ _mm_slli_si128(temp4, 4) ^ temp2;
}

static void key_256_assist_2(__m128i temp1, __m128i *temp3)
{
	__m128i temp2, temp4;

	temp2 = _mm_shuffle_epi32(_mm_aeskeygenassist_si128(temp1, 0), 0xaa);
	temp4 = _mm_slli_si128(*temp3, 4);
	*temp3 ^= temp4;
	temp4 = _mm_slli_si128(temp4, 4);
	*temp3 ^= temp4 ^ _mm_slli_si128(temp4, 4) ^ temp2;
}

void aes_hw_init(aes_hw_key *key)
{
	__m128i temp1, temp2;
	int i, j;

	key->key_enc[0] = temp1 = ((__m128i*)key->key)[0];
	key->key_enc[1] = temp2 = ((__m128i*)key->key)[1];
	for (i = 2, j = 1; i <= 12; i += 2, j <<= 1) {
		key_256_assist_1(&temp1, _mm_aeskeygenassist_si128(temp2, j));
		key->key_enc[i] = temp1;
		key_256_assist_2(temp1, &temp2);
		key->key_enc[i + 1] = temp2;
	}
	key_256_assist_1(&temp1, _mm_aeskeygenassist_si128(temp2, 0x40));
	key->key_enc[14] = temp1;

	key->key_dec[14] = key->key_enc[0];
	for (i = 1; i <= 13; i++)
		key->key_dec[14 - i] = _mm_aesimc_si128(key->key_enc[i]);
	key->key_dec[0] = key->key_enc[14];
}

void aes_hw_close(aes_hw_key *key)
{
	SecureZeroMemory(key, sizeof(aes_hw_key));
}

void aes_hw_encrypt(aes_hw_key *key, unsigned char *buffer)
{
	__m128i tmp;
	int i;

	tmp = *(__m128i*)buffer ^ key->key_enc[0];
	for (i = 1; i < 14; i++)
		tmp = _mm_aesenc_si128(tmp, key->key_enc[i]);
	*(__m128i*)buffer = _mm_aesenclast_si128(tmp, key->key_enc[14]);
}

void aes_hw_decrypt(aes_hw_key *key, unsigned char *buffer)
{
	__m128i tmp;
	int i;

	tmp = *(__m128i*)buffer ^ key->key_dec[0];
	for (i = 1; i < 14; i++)
		tmp = _mm_aesdec_si128(tmp, key->key_dec[i]);
	*(__m128i*)buffer = _mm_aesdeclast_si128(tmp, key->key_dec[14]);
}
